# frozen_string_literal: true
#
#   irb/version.rb - irb version definition file
#   	by Keiju ISHITSUKA(keiju@ishitsuka.com)
#

module IRB # :nodoc:
  VERSION = "1.15.0"
  @RELEASE_VERSION = VERSION
  @LAST_UPDATE_DATE = "2025-01-21"
end
